# -*- coding: utf-8 -*-

from flask import request, redirect, url_for

def redirect_to_self(clear=False):
	if clear:
		return redirect(url_for(request.endpoint))
	return redirect(url_for(request.endpoint, **request.view_args))

redirect_self = redirect_to_self

def redirect_to_next(clear=True, key='next'):
	if key in request.args:
		if clear:
			return redirect(request.args.get('next'))
		else:
			next = request.args.pop('next')
			return redirect(next, **request.view_args)

redirect_next = redirect_to_next


# -*- coding: utf-8 -*-

from flask import flash, jsonify

def handle_xhr_error(error):
	response = jsonify(error=error.message, code=error.status_code)
	response.status_code = error.status_code
	return response

class XHRError(Exception):
	status_code = 500

	def __init__(self, message, status_code=None):
		super(XHRError, self).__init__()
		self.message = message
		if status_code is not None:
			self.status_code = status_code


# -*- coding: utf-8 -*-

from flask import Flask, flash, g

from .. import EasyMode
from ..decorators import xhr_api, inject
from ..exceptions import XHRError
from ..mixins.model import CRUD, CRUDI, object_injected

em = EasyMode()

def create_app():
	app = Flask('easymode')

	@app.route('/')
	def index():
		return 'I am the index page.'

	@app.route('/xhr')
	@xhr_api()
	def xhr_endpoint():
		try:
			g.xhr.data['test'] = 'monkey'
			g.xhr.data['walrus'] = 'punch'
		except AttributeError: pass

		flash('This is a test message.')
		flash('This is a warning.', 'warning')
		flash('This is an error', 'error')
		flash('This is just some info', 'info')
		flash('This is another warning', 'warning')

	@app.route('/xhr-failure')
	@xhr_api()
	def xhr_failure():
		try:
			g.xhr.data['test'] = 'monkey'
		except AttributeError: pass

		raise XHRError('Disaster everywhere.')

	@app.route('/xhr-failure-with-code')
	@xhr_api()
	def xhr_failure_with_code():
		try:
			g.xhr.data['test'] = 'monkey'
		except AttributeError: pass

		raise XHRError('Disaster befalls the city', status_code=500)

	@app.route('/xhr-that-returns-something')
	@xhr_api()
	def xhr_that_returns_something():
		try:
			g.xhr.data['test'] = 'monkey'
		except AttributeError: pass

		return 'Here is some string that would never be returned if the XHR API were active.'

	@app.route('/xhr-that-allows-regular-http')
	@xhr_api(allow_http=True)
	def xhr_that_allows_regular_http():
		try:
			g.xhr.data['test'] = 'monkey'
		except AttributeError: pass

		flash('A message in a bottle.')
		
		return 'Here is some regular return stuff'

	@app.route('/inject/<injectable_class_slug_name>')
	@inject('injectable_class')
	def inject_test_class_slug_name():
		return 'I have been injected with %s' % g.injectable_class.slug_name

	@app.route('/inject-as-arg/<injectable_class_slug_name>')
	@inject('injectable_class', as_args=True)
	def inject_test_class_args(injectable_class):
		return 'I have been injected with %s' % injectable_class.slug_name

	@app.route('/inject-non-injectable/<non_injectable_class_slug_name>')
	@inject('non_injectable_class')
	def inject_the_noninjectable():
		return 'This will never happen because there will be an exception :('

	return app

def app_setup():
	em.init_app(app)

class InjectableClass(CRUDI):

	def __init__(self):
		self.slug_name = 'joe-slug'

class NonInjectableClass(CRUD): pass

@object_injected.connect_via(InjectableClass)
def injectable_injected(cls, conditions, **kwargs):
	for k, v in conditions:
		if k == 'slug_name' and v == 'joe-slug':
			return InjectableClass()

app = create_app()
app.config['TESTING'] = True
app.config['SECRET_KEY'] = '123454fdsafdsfdsfdsfds'


# -*- coding: utf-8 -*-

from nose.tools.nontrivial import with_setup
from nose.tools.trivial import assert_equals, assert_in, assert_true, assert_raises

from simplejson import loads

from . import app_setup, em, app, InjectableClass, NonInjectableClass

@with_setup(app_setup)
def test_xhr_api():	
	em.enable_xhr(app)

	with app.test_client() as c:
		r = c.get('/')
		assert_in('I am the index page', r.data)

		r = c.get('/xhr', headers=[('X-Requested-With', 'XMLHttpRequest')])

		py_data = loads(r.data)
		assert_in('data', py_data.keys())
		assert_in('messages', py_data.keys())

		assert_equals(len(py_data['data']), 2)
		assert_equals(len(py_data['messages']), 5)

		r = c.get('/xhr-failure')
		py_data = loads(r.data)
		assert_in('error', py_data.keys())

		r = c.get('/xhr-failure', headers=[('X-Requested-With', 'XMLHttpRequest')])
		py_data = loads(r.data)
		assert_in('error', py_data.keys())

		r = c.get('/xhr-failure-with-code', headers=[('X-Requested-With', 'XMLHttpRequest')])
		assert_equals(r.status_code, 500)
		py_data = loads(r.data)
		assert_in('error', py_data.keys())
		assert_in('code', py_data.keys())

		r = c.get('/xhr-that-allows-regular-http')
		py_data = loads(r.data)
		assert_in('test', py_data['data'].keys())
		assert_in('A message in a bottle.', py_data['messages'][0])

@with_setup(app_setup)
def test_xhr_api_off():
	with app.test_client() as c:
		r = c.get('/')
		assert_in('I am the index page.', r.data)

		r = c.get('/xhr')
		assert_in('403', r.data)

		r = c.get('/xhr', headers=[('X-Requested-With', 'XMLHttpRequest')])
		assert_in('403', r.data)

		r = c.get('/xhr-that-returns-something')
		assert_in('some string', r.data)

@with_setup(app_setup)
def test_inject():
	em.enable_injection(app)
	em.add_injectable(InjectableClass)

	with app.test_client() as c:
		r = c.get('/inject/joe-slug')
		assert_in('joe-slug', r.data)

		r = c.get('/inject-as-arg/joe-slug')
		assert_in('joe-slug', r.data)

		with assert_raises(RuntimeError):
			r = c.get('/inject-non-injectable/here-comes-an-error')

		em.add_injectable(NonInjectableClass)

		with assert_raises(RuntimeError):
			r = c.get('/inject-non-injectable/still-going-to-fail')

@with_setup(app_setup)
def test_inject_off():
	with app.test_client() as c:
		with assert_raises(Exception):
			r = c.get('/inject/only-bad-can-come-of-this')



# -*- coding: utf-8 -*-

import re

from flask import current_app, request, g, abort

from .exceptions import XHRError, handle_xhr_error

class EasyMode(object):

	_injectables = {}

	def __init__(self, app=None):
		if app is not None:
			self.init_app(app)

	def init_app(self, app):
		app.config['XHR_API_ENABLED'] = False
		app.config['XHR_API_ALLOW_HTTP'] = False
		app.config['VIEW_DI_ENABLED'] = False

	# as dirty as this looks, it's the best way to do it without using the decorator
	# because there is no built in function in flask yet to do it.
	def enable_xhr(self, app):
		app.config['XHR_API_ENABLED'] = True
		app.error_handler_spec.setdefault(None, {}).setdefault(None, []) \
		.append((XHRError, handle_xhr_error))

	def enable_injection(self, app):
		app.config['VIEW_DI_ENABLED'] = True
		
	def add_injectable(self, cls, alt=None):
		cls_name = alt or cls.__name__
		cls_name = re.sub(r'(?<=[a-z])[A-Z]|(?<!^)[A-Z](?=[a-z])', r"_\g<0>", cls_name).lower()
		EasyMode._injectables[cls_name] = cls


from setuptools import setup, find_packages

setup(
	name='flask-easymode',
	version='0.0.9',
	url='http://github.com/petermelias/flask-easymode',
	license='BSD',
	author='Peter M. Elias',
	author_email='petermelias@gmail.com',
	description='Make Flask development even easier',
	long_description=open('README.md').read(),
	packages=find_packages(),
	zip_safe=False,
	include_package_data=True,
	platforms='any',
	install_requires=[
		'Flask',
		'Werkzeug',
		'blinker',
		'simplejson'
	],
	classifiers=[
		'Environment :: Web Environment',
		'Intended Audience :: Developers',
		'License :: OSI Approved :: BSD License',
		'Operating System :: OS Independent',
		'Programming Language :: Python',
		'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
		'Topic :: Software Development :: Libraries :: Python Modules'
	],
)


# -*- coding: utf-8 -*-

from simplejson import dumps
from blinker import Namespace
from datetime import datetime

easy_signals = Namespace()

object_created = easy_signals.signal('object_created')
object_read = easy_signals.signal('object_read')
object_updated = easy_signals.signal('object_updated')
object_deleted = easy_signals.signal('object_deleted')
object_injected = easy_signals.signal('object_injected')

class Create(object):

	@classmethod
	def create(cls, **kwargs):
		o = cls()
		object_created.send(cls, o=o, **kwargs)
		return o

class Read(object):

	@classmethod
	def read(cls, **kwargs):
		r = object_read.send(cls, **kwargs)
		try:
			return r[0][1]
		except IndexError: pass

	@classmethod
	def read_many(cls, **kwargs):
		r = object_read.send(cls, _many=True, **kwargs)
		try:
			return r[0][1]
		except IndexError: pass

	@property
	def as_dict(self):
		d = {}
		for attr in self._readable:
			parts = attr.split('.')
			value = getattr(self, parts[0])
			for part in parts[1:]:
				if value is None:
					continue
				value = getattr(value, part)
			
			value = unicode(value)
		
			d[attr.replace('.', '_')] = value
		return d

	@property
	def as_json(self):
		return dumps(self.as_dict)

class Update(object):

	@classmethod
	def update(cls, o, **kwargs):
		r = object_updated.send(cls, o=o, **kwargs)
		try:
			return r[0][1]
		except IndexError: pass

class Delete(object):

	def delete(self):
		object_deleted.send(self)

class CRUD(Create, Read, Update, Delete): pass

class Injectable(object):

	@classmethod
	def load(cls, conditions, **kwargs):
		r = object_injected.send(cls, conditions=conditions, **kwargs)
		try:
			return r[0][1]
		except IndexError: pass

class CRUDI(CRUD, Injectable): pass


# -*- coding: utf-8 -*-

from flask import current_app, redirect, request, flash, url_for, g, \
	abort, get_flashed_messages, jsonify

from functools import wraps
from collections import namedtuple

from . import EasyMode
from .exceptions import XHRError

xhr = namedtuple('xhr', ('data', 'messages', 'html'))

class XHR(object):

	def __init__(self):
		self.html = ''
		self.data = {}
		self.messages = []

def xhr_api(allow_http=None):
	def _decorator(f):
		@wraps(f)
		def _wrapper(*args, **kwargs):
			if not current_app.config.get('XHR_API_ENABLED'):
				view_result = f(*args, **kwargs)
				if not view_result:
					abort(403)
				return view_result
			
			a = allow_http
			if a is None:
				a = current_app.config.get('XHR_API_ALLOW_HTTP')

			if not request.is_xhr and not a:
				raise XHRError('XHR endpoints must be called asynchronously.', status_code=500)

			g.xhr = XHR()
			f(*args, **kwargs)
			g.xhr.messages.extend(get_flashed_messages(with_categories=True))
			return jsonify(data=g.xhr.data, messages=g.xhr.messages, html=g.xhr.html)
		return _wrapper
	return _decorator


def inject(*classes, **options):
	def _decorator(f):
		@wraps(f)
		def _wrapper(*args, **kwargs):
			if not current_app.config.get('VIEW_DI_ENABLED'):
				return f(*args, **kwargs)
			
			# most to least specific to eliminate ambiguity when dealing with nested prefix class names
			classes_by_len = sorted(classes, key=len)
			classes_by_len.reverse()

			injections = {}

			def _param_to_cls_prop_pair(param):
				for cls_name in classes_by_len:
					r = param.replace(cls_name, '')
					if r != param:
						return (cls_name, r[1:])

			def _extract_injections(kvps):
				for k, v in kvps.iteritems():
					try:
						cls_name, prop_name = _param_to_cls_prop_pair(k)
					except TypeError:
						continue
					
					injections[cls_name]['params'].append(k)
					injections[cls_name]['conditions'].append((prop_name, v))

			for cls_name in classes:
				try:
					cls = EasyMode._injectables[cls_name]
				except KeyError:
					raise RuntimeError('Class "%s" has not been added as injectable. Use EasyMode.add_injectable(cls_name).' % cls_name)

				injections.setdefault(cls_name, {'class': cls, 'params': [], 'conditions': []})
			
			_extract_injections(kwargs)
			_extract_injections(request.form.to_dict())
			json = request.get_json(silent=True) or {}
			_extract_injections(json)
			_extract_injections(request.args)

			for cls_name, i in injections.iteritems():
				if cls_name in classes:
					try:
						o = i['class'].load(i['conditions'])
					except AttributeError:
						raise RuntimeError('To use %s with dependency injection, the class must either define a load(cls, conditions, **kwargs) interface or just inherit from the provided mixin.' % i['class'].__name__)
					
					if options.get('as_args'):
						kwargs[cls_name] = o
					else:
						setattr(g, cls_name, o)
	
					[kwargs.pop(p, None) for p in i['params']]

			return f(*args, **kwargs)
		return _wrapper
	return _decorator


# -*- coding: utf-8 -*-

from nose.tools.trivial import assert_equals, assert_true, assert_is_instance, assert_in

from ...mixins.model import CRUD, object_created, object_read, object_updated, \
	object_deleted


class CRUDMan(CRUD):

	_readable = ('name', 'catch_phrase')
	_updateable = ('name')


	def __init__(self):
		self.name = 'CRUD Man'
		self.catch_phrase = 'I am CRUD Man. Fear my power.'

def test_create():

	@object_created.connect
	def create_fxn(cls, o, **kwargs):
		o.touched_by_signal = True

	o = CRUDMan.create()

	assert_true(o.touched_by_signal)
	assert_is_instance(o, CRUDMan)

def test_read():

	@object_read.connect
	def reading_fxn(cls, **kwargs):
		assert_equals(cls, CRUDMan)
		assert_in('name', kwargs.keys())

		if kwargs.get('_many'):
			return 'Many things'
		return 'Something'

	r = CRUDMan.read(name='CRUD Man')
	assert_equals(r, 'Something')

	r = CRUDMan.read_many(name='CRUD Man')
	assert_equals(r, 'Many things')

	o = CRUDMan.create()
	o.as_dict
	o.as_json

def test_update():

	data_dict = {'name': 'CRUD Woman', 'catch_phrase': 'Cant change this'}

	@object_updated.connect
	def updating_fxn(cls, o, **kwargs):
		assert_equals(o.name, 'CRUD Man')
		assert_equals(o.catch_phrase, 'I am CRUD Man. Fear my power.')
		return o.as_dict

	o = CRUDMan.create()

	r = CRUDMan.update(o, **data_dict)
	assert_equals(r, {'name': 'CRUD Man', 'catch_phrase': 'I am CRUD Man. Fear my power.'})

def test_delete():

	o = CRUDMan.create()
	o.name = 'Dying CRUD Man'

	@object_deleted.connect
	def delete_fxn(o, **kwargs):
		assert_equals(o.name, 'Dying CRUD Man')

	o.delete()